@@ -2,20 +2,19 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 15.2.3.6-3-95
+es5id: 15.2.3.6-3-123
 description: >
-    Object.defineProperty - 'Attributes' is an Error object that uses
-    Object's [[Get]] method to access the 'configurable' property
-    (8.10.5 step 4.a)
+    Object.defineProperty - 'configurable' property in 'Attributes' is
+    the global object  (8.10.5 step 4.b)
 ---*/
 
 var obj = {};
 
-var errObj = new Error();
+var attr = {
+  configurable: this
+};
 
-errObj.configurable = true;
-
-Object.defineProperty(obj, "property", errObj);
+Object.defineProperty(obj, "property", attr);
 
 var beforeDeleted = obj.hasOwnProperty("property");
 
